@@ -14,13 +14,13 @@ use Digest::HMAC_SHA1 qw(hmac_sha1 hmac_sha1_hex);
 use MIME::Base64::URLSafe;
 use MIME::Base64;
 use String::HexConvert;
-
+use File::Slurp;
 
 
 use Exporter 'import';
 our @EXPORT = qw(DATABASE_RESET MAC_ERROR MAC_KEY_ERROR INTERNAL_ERROR SERVER_ERROR NO_UPDATE NO_DATA SUCCESSFUL MALWARE PHISHING);
 
-our $VERSION = '1.06';
+our $VERSION = '1.08';
 
 
 =head1 NAME
@@ -133,6 +133,8 @@ use constant {
 
 =over 4
 
+=back
+
 =head2 new()
 
 Create a Net::Google::SafeBrowsing2 object
@@ -175,14 +177,16 @@ The debug output maybe quite large and can slow down significantly the update an
 
 Optional. Set to 1 to show errors to STDOUT. 0 (disabled by default).
 
+=item perf
+
+Optional. Set to 1 to show performance information.
+
 =item version
 
 Optional. Google Safe Browsing version. 2.2 by default
 
 =back
 
-=back
-
 =cut
 
 sub new {
@@ -196,6 +200,7 @@ sub new {
 		errors		=> 0,
 		last_error	=> '',
 		mac			=> 0,
+		perf		=> 0,
 
 		%args,
 	};
@@ -217,6 +222,8 @@ sub new {
 
 =over 4
 
+=back
+
 =head2 update()
 
 Perform a database update.
@@ -265,8 +272,8 @@ sub update {
 
 	my $result = 0;
 
-
-	# too early to update? need to handle errors
+	# Too early to update?
+	my $start = time();
 	my $i = 0;
 	while ($i < scalar @lists) {
 		my $list = $lists[$i];
@@ -286,6 +293,7 @@ sub update {
 		$self->debug("Too early to update any list\n");
 		return NO_UPDATE;
 	}
+	$self->perf("OK to update check: " . (time() - $start) . "s\n");
 	
 	# MAC?
 	my $client_key = '';
@@ -309,8 +317,10 @@ sub update {
 	my $body = '';
 	foreach my $list (@lists) {
 		# Report existng chunks
+		$start = time();
 		my $a_range = $self->create_range(numbers => [$self->{storage}->get_add_chunks_nums(list => $list)]);
 		my $s_range = $self->create_range(numbers => [$self->{storage}->get_sub_chunks_nums(list => $list)]);
+		$self->perf("Create add and sub ranges: " . (time() - $start) . "s\n");
 	
 		my $chunks_list = '';
 		if ($a_range ne '') {
@@ -326,12 +336,14 @@ sub update {
 		$body .= "\n";
 	}
 
-
+	my $start_req = time();
 	my $res = $ua->post($url, Content =>  $body);
+	$self->perf("$body\n");
 
 # 	$self->debug($res->request->as_string . "\n" . $res->as_string . "\n");
-	$self->debug($res->request->as_string . "\n");
-	$self->debug($res->as_string . "\n");
+	$self->debug($res->request->as_string . "\n") if ($self->{debug});
+	$self->debug($res->as_string . "\n") if ($self->{debug});
+	my $duration_req = time() - $start_req;
 
 	if (! $res->is_success) {
 		$self->error("Request failed\n");
@@ -347,6 +359,9 @@ sub update {
 	my $wait = 0;
 
 	my @redirections = ();
+	my $del_add_duration = 0;
+	my $del_sub_duration = 0;
+	my $add_range_info = '';
 
 	my @lines = split/\s/, $res->decoded_content;
 	$list = '';
@@ -371,19 +386,24 @@ sub update {
 		elsif ($line =~ /ad:(\S+)$/) {
 			$self->debug("Delete Add Chunks: $1\n");
 
+			my $del_add_start = time();
+			$add_range_info = $1 . " $list";
 			my @nums = $self->expand_range(range => $1);
 			$self->{storage}->delete_add_ckunks(chunknums => [@nums], list => $list);
 
 			# Delete full hash as well
 			$self->{storage}->delete_full_hashes(chunknums => [@nums], list => $list);
+			$del_add_duration = time() - $del_add_start;
 
 			$result = 1;
 		}
 		elsif ($line =~ /sd:(\S+)$/) {
 			$self->debug("Delete Sub Chunks: $1\n");
 
+			my $del_sub_start = time();
 			my @nums = $self->expand_range(range => $1);
 			$self->{storage}->delete_sub_ckunks(chunknums => [@nums], list => $list);
+			$del_add_duration = time() - $del_sub_start;
 
 			$result = 1;
 		}
@@ -416,11 +436,13 @@ sub update {
 		}
 	}
 	$self->debug("\n");
+	$self->perf("Handle first request: " . (time() - $last_update) . "s (POST: ${duration_req}s, DEL add: ${del_add_duration}s, DEL sub: ${del_sub_duration}s, ADD range: ${add_range_info})\n");
 
 	$result = 1 if (scalar @redirections > 0);
 
-
+	$self->perf("Parse redirections: ");
 	foreach my $data (@redirections) {
+		$start = time();
 		my $redirection = $data->[0];
 		$list = $data->[1];
 		my $hmac = $data->[2];
@@ -447,69 +469,18 @@ sub update {
 
 			return MAC_ERROR;
 		}
-	
-		my $chunk_num = 0;
-		my $hash_length = 0;
-		my $chunk_length = 0;
-	
-		while (length $data > 0) {
-	# 		print "Length 1: ", length $data, "\n"; # 58748
-	
-			my $type = substr($data, 0, 2, ''); # s:34321:4:137
-	# 		print "Length 1.5: ", length $data, "\n"; # 58746 -2
-	
-			if ($data  =~ /^(\d+):(\d+):(\d+)\n/sgi) {
-				$chunk_num = $1;
-				$hash_length = $2;
-				$chunk_length = $3;
-	
-				# shorten data
-				substr($data, 0, length($chunk_num) + length($hash_length) + length($chunk_length) + 3, '');
-	# 			print "Remove ", length($chunk_num) + length($hash_length) + length($chunk_length) + 3, "\n";
-	# 			print "Length 2: ", length $data, "\n"; # 58741 -5
-	
-				my $encoded = substr($data, 0, $chunk_length, '');
-	# 			print "Length 3: ", length $data, "\n"; # 58604 -137
-	
-				if ($type eq 's:') {
-					my @chunks = $self->parse_s(value => $encoded, hash_length => $hash_length);
 
-					$self->{storage}->add_chunks(type => 's', chunknum => $chunk_num, chunks => [@chunks], list => $list); # Must happen all at once => not 100% sure
-				}
-				elsif ($type eq 'a:') {
-					my @chunks = $self->parse_a(value => $encoded, hash_length => $hash_length);
-					$self->{storage}->add_chunks(type => 'a', chunknum => $chunk_num, chunks => [@chunks], list => $list); # Must happen all at once => not 100% sure
-				}
-				else {
-					$self->error("Incorrect chunk type: $type, should be a: or s:\n");
-
-					foreach my $list (@lists) {
-						$self->update_error('time' => $last_update, list => $list);
-					}
-					return INTERNAL_ERROR;# failed
-				}
-	
-				$self->debug("$type$chunk_num:$hash_length:$chunk_length OK\n");
-			
+		my $result = $self->parse_data(data => $data, list => $list);
+		if ($result != SUCCESSFUL) {
+			foreach my $list (@lists) {
+				$self->update_error('time' => $last_update, list => $list);
 			}
-# 			elsif ($data =~ /e:pleaserekey/ && $mac) { # May neverr happen here?
-# 				$self->Debug("MAC key has been expired (redirection)\n");
-# 	
-# 				$self->{storage}->delete_mac_keys();
-# 				return $self->update(list => $list, force => $force, mac => $mac);
-# 			}
-			else {
-				$self->error("could not parse header\n");
 
-				foreach my $list (@lists) {
-					$self->update_error('time' => $last_update, list => $list);
-				}
-				return INTERNAL_ERROR;# failed
-			}
+			return $result;
 		}
-
-	# 	last;
+		$self->perf((time() - $start) . "s ");
 	}
+	$self->perf("\n");
 
 	foreach my $list (@lists) {
 		$self->debug("List update: $last_update $wait $list\n");
@@ -519,6 +490,43 @@ sub update {
 	return $result; # ok
 }
 
+=head2 import_chunks()
+
+Import add and sub chunks from a file.
+
+  my $result = $gsb->import_chunks(list => MALWARE, file => 'malware.dat');
+
+Return the status of the import: INTERNAL_ERROR or SUCCESSFUL.
+
+This function should be used to initialize an empty back-end storage.
+
+
+Arguments
+
+=over 4
+
+=item list
+
+Required. Google list to use.
+
+=item file
+
+Required. File that contains the list of chunks. This file can be created with the C<export> function inherited from C<Net::Google::SafeBrowsing2::DBI>.
+
+=back
+
+=cut
+
+sub import_chunks {
+	my ($self, %args) 	= @_;
+	my $list 			= $args{list}		|| '';
+	my $file			= $args{file}		|| "$list.dat";
+
+	my $data = read_file($file, { binmode => ':raw' });
+
+	return $self->parse_data(data => $data, list => $list);
+
+}
 
 =head2 lookup()
 
@@ -625,14 +633,14 @@ sub last_error {
 
 =pod
 
-=back
-
 =head1 PRIVATE FUNCTIONS
 
 These functions are not intended to be used externally.
 
 =over 4
 
+=back
+
 =head2 lookup_suffix()
 
 Lookup a host prefix.
@@ -669,7 +677,7 @@ sub lookup_suffix {
 		foreach my $full_hash (@full_hashes) {
 			foreach my $hash (@hashes) {
 				if ($hash eq $full_hash && defined first { $add_chunk->{list} eq $_ } @$lists) {
-					$self->debug("Full hash was found in storage\n");
+					$self->debug("Full hash was found in storage: " . $self->hex_to_ascii($hash) . "\n");
 					$found = $add_chunk->{list};
 					last;
 				}
@@ -699,7 +707,7 @@ sub lookup_suffix {
 		if (defined $hash && defined $list) {
 # 			$self->debug($self->hex_to_ascii($hash->{hash}) . " eq " . $self->hex_to_ascii($full_hash) . "\n\n");
 
-			$self->debug("Match\n");
+			$self->debug("Match: " . $self->hex_to_ascii($full_hash)  . "\n");
 
 			return $hash->{list};
 		}
@@ -1037,6 +1045,67 @@ sub ua {
 }
 
 
+=head2 parse_s()
+
+Parse data from a rediration (add asnd sub chunk information).
+
+=cut
+
+sub parse_data {
+	my ($self, %args) 	= @_;
+	my $data			= $args{data}		 || '';
+	my $list  			= $args{list}		 || '';
+
+	my $chunk_num = 0;
+	my $hash_length = 0;
+	my $chunk_length = 0;
+
+	while (length $data > 0) {
+	# 		print "Length 1: ", length $data, "\n"; # 58748
+	
+			my $type = substr($data, 0, 2, ''); # s:34321:4:137
+	# 		print "Length 1.5: ", length $data, "\n"; # 58746 -2
+	
+			if ($data  =~ /^(\d+):(\d+):(\d+)\n/sgi) {
+				$chunk_num = $1;
+				$hash_length = $2;
+				$chunk_length = $3;
+	
+				# shorten data
+				substr($data, 0, length($chunk_num) + length($hash_length) + length($chunk_length) + 3, '');
+	# 			print "Remove ", length($chunk_num) + length($hash_length) + length($chunk_length) + 3, "\n";
+	# 			print "Length 2: ", length $data, "\n"; # 58741 -5
+	
+				my $encoded = substr($data, 0, $chunk_length, '');
+	# 			print "Length 3: ", length $data, "\n"; # 58604 -137
+	
+				if ($type eq 's:') {
+					my @chunks = $self->parse_s(value => $encoded, hash_length => $hash_length);
+
+					$self->{storage}->add_chunks(type => 's', chunknum => $chunk_num, chunks => [@chunks], list => $list); # Must happen all at once => not 100% sure
+				}
+				elsif ($type eq 'a:') {
+					my @chunks = $self->parse_a(value => $encoded, hash_length => $hash_length);
+					$self->{storage}->add_chunks(type => 'a', chunknum => $chunk_num, chunks => [@chunks], list => $list); # Must happen all at once => not 100% sure
+				}
+				else {
+					$self->error("Incorrect chunk type: $type, should be a: or s:\n");
+					return INTERNAL_ERROR;# failed
+				}
+	
+				$self->debug("$type$chunk_num:$hash_length:$chunk_length OK\n");
+			
+			}
+			else {
+				$self->error("could not parse header\n");
+				return INTERNAL_ERROR;# failed
+			}
+		}
+
+	return SUCCESSFUL;
+}
+
+
 =head2 parse_s()
 
 Parse s chunks information for a database update.
@@ -1046,7 +1115,7 @@ Parse s chunks information for a database update.
 sub parse_s {
 	my ($self, %args) 	= @_;
 	my $value 			= $args{value}			|| return ();
-	my $hash_length 	= $args{hash_length}	|| 4;;
+	my $hash_length 	= $args{hash_length}	|| 4;
 
 	my @data = ();
 
@@ -1208,6 +1277,19 @@ sub error {
 	$self->{last_error} = $message;
 }
 
+
+=head2 error()
+
+Print performance message.
+
+=cut
+
+sub perf {
+	my ($self, $message) = @_;
+
+	print $message if ($self->{perf} > 0);
+}
+
 =head2 canonical_domain_suffixes()
 
 Find all suffixes for a domain.
@@ -1359,7 +1441,8 @@ sub canonical_uri {
 
 	# canonial does not handle ../ 
 # 	$self->debug("\t$escape\n");
-	while($escape =~ s/([^\/])\/([^\/]+)\/\.\.([\/?].*)?$/$1$3/gi) {  }
+	while($escape =~ s/([^\/])\/([^\/]+)\/\.\.([\/?].*)$/$1$3/gi) {  }
+	while($escape =~ s/([^\/])\/([^\/]+)\/\.\.$/$1/gi) {  }
 
 	# May have removed ending /
 # 	$self->debug("\t$escape\n");
@@ -1396,7 +1479,8 @@ sub canonical_uri {
 # 	$self->debug("2. $url => $escape\n");
 
 	# Fix over escaping
-	while($escape =~ s/^([^?]+)%%(%.*)?$/$1%25%25$2/sgi) { }
+	while($escape =~ s/^([^?]+)%%(%.*)$/$1%25%25$2/sgi) { }
+	while($escape =~ s/^([^?]+)%%/$1%25%25/sgi) { }
 
 	# URI has issues with % in domains, it gets the host wrong
 
@@ -1453,6 +1537,7 @@ sub full_hashes {
 	foreach my $url (@urls) {
 # 		$self->debug("$url\n");
 		push(@hashes, sha256($url));
+# 		$self->debug("$url " . $self->hex_to_ascii(sha256($url)) . "\n");
 	}
 
 	return @hashes;
@@ -1681,15 +1766,62 @@ sub expand_range {
 }
 
 
-=back
-
 =head1 CHANGELOG
 
 =over 4
 
-=item 0.2
+=item 1.07
 
-Add support for Message Authentication Code (MAC)
+Add C<import_chunks()> feature to import add chunks and sub chunks from a file.
+
+=item 1.05
+
+No code change. Move C<local_lookup> to PRIVATE FUNCTIONS to avoid confusions.
+
+=item 1.04
+
+Introduce L<Net::Google::SafeBrowsing2::Lookup>. Remind people that Google Safe Browsing v1 has been deprecated by Google.
+
+=item 1.03
+
+The source code is available on github at L<https://github.com/juliensobrier/Net-Google-SafeBrowsing2>.
+
+=item 1.02
+
+Fix uninitialized $self->{errors} variable
+
+=item 1.01
+
+Use String::HexConvert for faster hex_to_ascii.
+
+=item 1.0
+
+Separate the error output from the debug output.
+
+=item 0.9
+
+Fix bug with local whitelisting (sub chunks). Fix the parsing of full hashes.
+
+=item 0.8
+
+Reduce the number of full hash requests.
+
+=item 0.7
+
+Add local_lookup to perform a lookup against the local database only. This function should be used for debugging purpose only.
+Small code optimizations.
+
+=item 0.6
+
+Handle local database reset.
+
+=item 0.5
+
+Update documentation.
+
+=item 0.4
+
+Speed update the database update. The first update went down from 20 minutes to 20 minutes.
 
 =item 0.3
 
@@ -1699,54 +1831,9 @@ Remove dependency on Switch (thanks to Curtis Jewel).
 
 Fix value of FULL_HASH_TIME.
 
-=item 0.4
+=item 0.2
 
-Speed update the database update. The first update went down from 20 minutes to 20 minutes.
-
-=item 0.5
-
-Update documentation.
-
-=item 0.6
-
-Handle local database reset.
-
-=item 0.7
-
-Add local_lookup to perform a lookup against the local database only. This function should be used for debugging purpose only.
-Small code optimizations.
-
-=item 0.8
-
-Reduce the number of full hash requests.
-
-=item 0.9
-
-Fix bug with local whitelisting (sub chunks). Fix the parsing of full hashes.
-
-=item 1.0
-
-Separate the error output from the debug output.
-
-=item 1.01
-
-Use String::HexConvert for faster hex_to_ascii.
-
-=item 1.02
-
-Fix uninitialized $self->{errors} variable
-
-=item 1.03
-
-The source code is available on github at L<https://github.com/juliensobrier/Net-Google-SafeBrowsing2>.
-
-=item 1.04
-
-Introduce L<Net::Google::SafeBrowsing2::Lookup>. Remind people that Google Safe Browsing v1 has been deprecated by Google.
-
-=item 1.05
-
-No code change. Move C<local_lookup> to PRIVATE FUNCTIONS to avoid confusions.
+Add support for Message Authentication Code (MAC)
 
 =back
 
@@ -1773,4 +1860,4 @@ at your option, any later version of Perl 5 you may have available.
 
 =cut
 
-1;
\ No newline at end of file
+1;
